The virtual environment is now created.

You may activate it by sourcing the appropriate activation script in the
[green]bin[/] directory, and deactivate by invoking `ruyi-deactivate`.
{%- if sysroot %}

A fresh sysroot/prefix is also provisioned in the virtual environment.
It is available at the following path:

    [green]{{ sysroot }}[/]
{%- endif %}

The virtual environment also comes with ready-made CMake toolchain file
and Meson cross file. Check the virtual environment root for those;
comments in the files contain usage instructions.
